from .utils import BaseEstimator
import numpy as np

"""
This module includes code from pakcage scikit-survival: https://github.com/sebp/scikit-survival
    - Source File: sksurv/metrics.py
    - License: GPLv3
    - Modifications: Optimized for metrics computing precision, 
                     handling more general cases: e.g. y_test is outside the scope of y_train; y_train and y_test contains zero.  
"""

def _integrated_brier_score(G, event_test, y_test, estimates, times):
    """ Helper function for osst score function: compute the IBS score 
    
    """
    brier_scores = _brier_score(G, event_test, y_test, estimates, times)
    ibs = np.sum([ brier_scores[i] * (times[i+1] - times[i]) for i in range(len(times) - 1)]) / (times[-1] - times[0])
    return ibs 


def _brier_score(G, event_test, y_test, estimates, times):
    """
    Helper function for osst score function: compute the time-dependent brier scores
    """
    n_times = times.shape[0]
    ipcw_times = G.predict_ipcw(times)
    ipcw_y_test = G.predict_ipcw(y_test)


    brier_scores = np.zeros(n_times)
    for i in range(n_times):
        S_hat = estimates[:, i]
        brier_scores[i] = np.mean(np.square(S_hat) * ((y_test <= times[i]) & event_test).astype(int) * ipcw_y_test + 
                                  np.square(1.0 - S_hat) * (y_test > times[i]).astype(int) * ipcw_times[i])
        
    return brier_scores

def integrated_brier_score(event_train, y_train, event_test, y_test, estimates, times):
    """ The Integrated Brier Score (IBS)

    Parameters
    ----------
    event_train : array-like, shape = (n_train_samples,) or (n_train_samples, 1)
                  column containing event indicator for each training sample

    y_train : array-like, shape = (n_train_samples,) or (n_train_samples, 1)
              column containing the observed time for each training sample


    event_test : array-like, shape = (n_test_samples,) or (n_train_samples, 1)
                 column containing event indicator for each testing sample

    y_test : array-like, shape = (n_test_samples,) or (n_train_samples, 1)
              column containing the observed time for each testing sample

    estimate : array-like, shape = (n_test_samples, n_times)
        Estimated survival probability of remaining event-free at time points
        specified by `times`. The value of ``estimates[i][j]`` must correspond to
        the estimated probability of remaining event-free up to the time point
        the predicted survival proability of sample `i` at ``times[j]``. 
        Typically, estimated probabilities are obtained via the
        survival function returned by an estimator's ``predict_survival_function`` method.

    times : array-like, shape = (n_times,) 
        The time points for which to estimate the time-dependent brier scores.
        Values must be sorted in ascending order.


    tied_tol : float, optional, default: 1e-8
        The tolerance value for considering ties.
        If the absolute difference between risk scores is smaller
        or equal than `tied_tol`, risk scores are considered tied.

    Returns
    -------

    ibs : float
        The integrated Brier score.


    """
    if len(event_train.shape) > 1:
        event_train = event_train.values.reshape(-1)
    if len(y_train.shape) > 1:
        y_train = y_train.values.reshape(-1)
    if len(event_test.shape) > 1:
        event_test = event_test.values.reshape(-1)
    if len(y_test.shape) > 1:
        y_test = y_test.values.reshape(-1)

    brier_scores = brier_score(event_train, y_train, event_test, y_test, estimates, times)
    ibs = np.sum([ brier_scores[i] * (times[i+1] - times[i]) for i in range(len(times) - 1)]) / (times[-1] - times[0])
    return ibs


def brier_score(event_train, y_train, event_test, y_test, estimates, times):
    """ Time-dependent Brier score

    Parameters
    ----------
    event_train : array-like, shape = (n_train_samples,)
                  column containing event indicator for each training sample

    y_train : array-like, shape = (n_train_samples,)
              column containing the observed time for each training sample


    event_test : array-like, shape = (n_test_samples,)
                 column containing event indicator for each testing sample

    y_test : array-like, shape = (n_test_samples,)
              column containing the observed time for each testing sample

    Estimated survival probability of remaining event-free at time points
        specified by `times`. The value of ``estimates[i][j]`` must correspond to
        the estimated probability of remaining event-free up to the time point
        the predicted survival proability of sample `i` at ``times[j]``. 
        Typically, estimated probabilities are obtained via the survival function 
        returned by an estimator's ``predict_survival_function`` method.

    times : array-like, shape = (n_times,)
        The time points for which to estimate the Brier score.
        Values must be within the range of follow-up times of
        the test data `survival_test`.

    Returns
    -------
    brier_scores : array , shape = (n_times,)
        Values of the brier score.
    """
    n_times = times.shape[0]
    assert n_times == estimates.shape[1]
    G = BaseEstimator(event=event_train, y=y_train, reverse=True)
    ipcw_times = G.predict_ipcw(times)
    ipcw_y_test = G.predict_ipcw(y_test)


    brier_scores = np.zeros(n_times)
    for i in range(n_times):
        S_hat = estimates[:, i]
        brier_scores[i] = np.mean(np.square(S_hat) * ((y_test <= times[i]) & event_test).astype(int) * ipcw_y_test
                                 + np.square(1.0 - S_hat) * (y_test > times[i]).astype(int) * ipcw_times[i])
        
    return brier_scores



def _find_all_comparable_pairs(event, time, order):
    """ Helper function for _weighted_c_index: find all comparable pairs
    """
    n = len(time)
    i = 0
    while i < n - 1:
        time_i = time[order[i]]
        end = i + 1
        while end < n and event[order[end]] == time_i:
            end += 1
        
        censored_at_same_time = ~event[order[i:end]]
        for j in range(i, end):
            if event[order[j]] > 0:
                comparable_idx = np.zeros(n, dtype=bool)
                comparable_idx[end:] = True 
                comparable_idx[i:end] = censored_at_same_time
                
                yield (j, comparable_idx)
        
        i = end



def _weighted_c_index(event, time, estimates, threshold_mapping, weights, tied_tol=1e-8):
    """ Helper function for harrell_c_index and uno_c_index 
    """
    order = np.argsort(time)
    num_concordant = 0
    num_tied = 0
    num_comparable = 0
    numerator = 0.0
    denominator = 0.0
    for i, comparable_idx in _find_all_comparable_pairs(event, time, order):
        est_i = estimates[order[i]][threshold_mapping[order[i]]]
        event_i = event[order[i]]
        w_i = weights[order[i]]
        est = estimates[order[comparable_idx]][:, threshold_mapping[order[i]]]
        assert event_i > 0

        concordant = est > est_i
        ties = np.absolute(est - est_i) <= tied_tol
        n_ties = ties.sum()
        n_con = concordant[~ties].sum()
        n_comparable = comparable_idx.sum()

        numerator += w_i * n_con + 0.5 * w_i * n_ties
        denominator += w_i * n_comparable 

        num_comparable += n_comparable 
        num_concordant += n_con
        num_tied += n_ties

    
    if num_comparable < 1:
        raise Exception("No comparable pairs available.")
    
    return numerator / denominator, num_concordant, num_tied, num_comparable





def harrell_c_index(event, y, estimates, times):
    """Concordance index for right-censored data
    
    Parameters
    ----------
    event : array-like, shape = (n_samples,) or (n_samples, 1)
                  column containing event indicator for each testing sample

    y : array-like, shape = (n_samples,) or (n_samples, 1)
              column containing the observed time for each testing sample

    estimate : array-like, shape = (n_samples, n_times)
        Estimated survival probability of remaining event-free at time points
        specified by `times`. The value of ``estimates[i][j]`` must correspond to
        the estimated probability of remaining event-free up to the time point
        the predicted survival proability of sample `i` at ``times[j]``. 
        Typically, estimated probabilities are obtained via the
        survival function returned by an estimator's ``predict_survival_function`` method.

    times : array-like, shape = (n_times,) 
        The time points for which to estimate the harrell's c-index score.
        Values must be sorted in ascending order.


    Returns
    -------
    cindex : float
        Concordance index

    num_concordant : int
        Number of concordant pairs

    num_tied : int
        Number of tied pairs

    num_comparable : int
        Number of comparable pairs

        
    """
    if len(event.shape) > 1:
            event = event.values.reshape(-1)
    if len(y.shape) > 1:
        y = y.values.reshape(-1)

    if estimates.shape[0] != len(event) or estimates.shape[0] != len(y):
        raise ValueError("Estimates should have same row number as length of event and y")
    if estimates.shape[1] != len(times):
        raise ValueError("Estimates should have same column number as number of time thresholds ")
    weights = np.ones_like(y)
    time_threshold_mapping = np.searchsorted(times, y, side='left')
    assert (times[time_threshold_mapping] == y).all()

    return _weighted_c_index(event, y, estimates, time_threshold_mapping, weights)
    
def uno_c_index(event_train, y_train, event_test, y_test, estimates, times):
    """ Concordance index for right-censored data based on inverse probability of censoring weights.

    Parameters
    ----------
    event_train : array-like, shape = (n_train_samples,) or (n_train_samples, 1)
                  column containing event indicator for each training sample

    y_train : array-like, shape = (n_train_samples,) or (n_train_samples, 1)
              column containing the observed time for each training sample


    event_test : array-like, shape = (n_test_samples,) or (n_train_samples, 1)
                 column containing event indicator for each testing sample

    y_test : array-like, shape = (n_test_samples,) or (n_train_samples, 1)
              column containing the observed time for each testing sample

    estimate : array-like, shape = (n_test_samples, n_times)
        Estimated survival probability of remaining event-free at time points
        specified by `times`. The value of ``estimates[i][j]`` must correspond to
        the estimated probability of remaining event-free up to the time point
        the predicted survival proability of sample `i` at ``times[j]``. 
        Typically, estimated probabilities are obtained via the
        survival function returned by an estimator's ``predict_survival_function`` method.

    times : array-like, shape = (n_times,) 
        The time points for which to estimate the uno's c-index score.
        Values must be sorted in ascending order.


    Returns
    -------
    cindex : float
        Concordance index

    num_concordant : int
        Number of concordant pairs

    num_tied : int
        Number of tied pairs

    num_comparable : int
        Number of comparable pairs

    """
    if len(event_train.shape) > 1:
        event_train = event_train.values.reshape(-1)
    if len(y_train.shape) > 1:
        y_train = y_train.values.reshape(-1)
    if len(event_test.shape) > 1:
        event_test = event_test.values.reshape(-1)
    if len(y_test.shape) > 1:
        y_test = y_test.values.reshape(-1)

    if estimates.shape[0] != len(event_test) or estimates.shape[0] != len(y_test):
        raise ValueError("Estimates should have same row number as length of event_test and y_test")
    if estimates.shape[1] != len(times):
        raise ValueError("Estimates should have same column number as number of time thresholds ")
    
    # compute ipcw
    G = BaseEstimator(event=event_train, y=y_train, reverse=True)
    weights = np.square(G.predict_ipcw(y_test))

    time_threshold_mapping = np.searchsorted(times, y_test, side='left')
    assert (times[time_threshold_mapping] == y_test).all()
    return _weighted_c_index(event_test, y_test, estimates, time_threshold_mapping, weights)


def cumulative_dynamic_auc(event_train, y_train, event_test, y_test, estimates, times, tied_tol=1e-8):
    """ cumulative/dynamic AUC metric for right-censored time-to-event data.

    Parameters
    ----------
    event_train : array-like, shape = (n_train_samples,) or (n_train_samples, 1)
                  column containing event indicator for each training sample

    y_train : array-like, shape = (n_train_samples,) or (n_train_samples, 1)
              column containing the observed time for each training sample


    event_test : array-like, shape = (n_test_samples,) or (n_train_samples, 1)
                 column containing event indicator for each testing sample

    y_test : array-like, shape = (n_test_samples,) or (n_train_samples, 1)
              column containing the observed time for each testing sample

    estimate : array-like, shape = (n_test_samples, n_times)
        Estimated survival probability of remaining event-free at time points
        specified by `times`. The value of ``estimates[i][j]`` must correspond to
        the estimated probability of remaining event-free up to the time point
        the predicted survival proability of sample `i` at ``times[j]``. 
        Typically, estimated probabilities are obtained via the
        survival function returned by an estimator's ``predict_survival_function`` method.

    times : array-like, shape = (n_times,) 
        The time points for which to estimate the time-dependent AUC score.
        Values must be sorted in ascending order.


    tied_tol : float, optional, default: 1e-8
        The tolerance value for considering ties.
        If the absolute difference between risk scores is smaller
        or equal than `tied_tol`, risk scores are considered tied.

    Returns
    -------
    mean_auc : float
        Summary measure referring to the mean cumulative/dynamic AUC
        over the specified time range in `times`.

    auc scores : array, shape = (n_times',)
        The valid time-dependent AUC scores (evaluated at `times`). 
        n_times' <= n_times

    times': array, shape = (n_times',)
        The valid time thresholds to compute time-dependent AUC score
    

    """
    # input shape check
    if len(event_train.shape) > 1:
        event_train = event_train.values.reshape(-1)
    if len(y_train.shape) > 1:
        y_train = y_train.values.reshape(-1)
    if len(event_test.shape) > 1:
        event_test = event_test.values.reshape(-1)
    if len(y_test.shape) > 1:
        y_test = y_test.values.reshape(-1)
    
    assert len(event_test) == estimates.shape[0] and len(y_test) == estimates.shape[0]
    
    assert estimates.shape[1] == len(times)

    n_samples, n_times = estimates.shape

    # compute ipcw
    cens = BaseEstimator(event_train, y_train, True)
    ipcw_y_test = cens.predict_ipcw(y_test)

    # convert event_test, y_test, times, ipcw_y_test to (n_samples, n_times) shape
    event_test = np.broadcast_to(event_test[:, np.newaxis], (n_samples, n_times))
    y_test = np.broadcast_to(y_test[:, np.newaxis], (n_samples, n_times))
    times_2d = np.broadcast_to(times, (n_samples, n_times))
    ipcw_y_test = np.broadcast_to(ipcw_y_test[:, np.newaxis], (n_samples, n_times))

    # sort each column by estimated survival probability
    o = np.argsort(estimates, axis=0)
    estimates = np.take_along_axis(estimates, o, axis=0)
    event_test = np.take_along_axis(event_test, o, axis=0)
    y_test = np.take_along_axis(y_test, o, axis=0)
    ipcw_y_test = np.take_along_axis(ipcw_y_test, o, axis=0)

    # compute time-dependent AUC score
    is_case = (y_test <= times_2d) & event_test # each col indicates which samples are in case
    is_control = y_test > times_2d

    estimate_diff = np.concatenate((np.broadcast_to(np.inf, (1, n_times)), estimates))
    is_tied = np.absolute(np.diff(estimate_diff, axis=0)) <= tied_tol

    cumsum_tp = np.cumsum(is_case * ipcw_y_test, axis=0)
    cumsum_fp = np.cumsum(is_control, axis=0)
    # handle when no control / no case:
    valid_time_idx = ~ ((cumsum_tp[-1] == 0 ) | (cumsum_fp[-1] == 0) )
    cumsum_tp = cumsum_tp[:, valid_time_idx]
    cumsum_fp = cumsum_fp[:, valid_time_idx]
    is_tied = is_tied[:, valid_time_idx]
    times = times[valid_time_idx]

    true_pos = cumsum_tp / cumsum_tp[-1]
    false_pos = cumsum_fp / cumsum_fp[-1] 

    scores = np.empty(valid_time_idx.sum(), dtype=float)  

    for j in range(len(times)):
        tp = true_pos[:, j]
        fp = false_pos[:, j]
        mask = is_tied[:, j]

        idx = np.flatnonzero(mask) - 1
        # only keep the last estimate for tied risk scores
        if idx[0] < 0:
            # handle case when first estimate is tied
            idx = idx[1:]
        tp_no_ties = np.delete(tp, idx)
        fp_no_ties = np.delete(fp, idx)
        # Add an extra threshold position
        # to make sure that the curve starts at (0, 0)
        tp_no_ties = np.r_[0, tp_no_ties]
        fp_no_ties = np.r_[0, fp_no_ties]
    
        scores[j] = np.trapz(tp_no_ties, fp_no_ties)


    if n_times == 1:
        mean_auc = scores[0]
    else:
        surv = BaseEstimator(event_test[:, 0], y_test[:, 0])
        s_times = surv.predict_survival_prob(times)
        # compute integral of AUC over survival function
        d = -np.diff(np.r_[1.0, s_times])
        integral = (scores * d).sum()
        mean_auc = integral / (1.0 - s_times[-1])

    return mean_auc, scores, times


def compute_ibs_per_sample(event_train, y_train, event_test, y_test, estimates, times):
    """ Compute IBS score for each sample
    """
    if len(event_train.shape) > 1:
        event_train = event_train.values.reshape(-1)
    if len(y_train.shape) > 1:
        y_train = y_train.values.reshape(-1)
    if len(event_test.shape) > 1:
        event_test = event_test.values.reshape(-1)
    if len(y_test.shape) > 1:
        y_test = y_test.values.reshape(-1)

    assert len(event_test) == estimates.shape[0] and len(y_test) == estimates.shape[0]
    
    assert estimates.shape[1] == len(times) 

    n_samples, n_times = estimates.shape

    # compute ipcw
    cens = BaseEstimator(event_train, y_train, True)
    ipcw_times = cens.predict_ipcw(times)
    ipcw_y_test = cens.predict_ipcw(y_test)
    
    # compute ibs loss per sample
    ibs = np.zeros(n_samples, dtype=float)

    for i in range(n_times - 1):
        ibs += (np.square(estimates[:, i]) * ((y_test <= times[i]) & event_test).astype(int) * ipcw_y_test + \
                np.square(1.0 - estimates[:, i]) * (y_test > times[i]).astype(int) * ipcw_times[i]) * (times[i+1] - times[i])

    return ibs / (times[-1] - times[0]) / n_samples

    



    


import operator

from math import ceil, log
from numpy import array, argsort, hstack, dtype, issubdtype, isnan, nan, int64, int32, float64, float32
from pandas import isnull
from sortedcontainers import SortedList, SortedSet, SortedDict
from sklearn.ensemble import RandomForestClassifier

# Converts arbitrary datasets to binary values
class Encoder:
    def __is_numeric__(value):
        return issubdtype(dtype(type(value)), dtype(int)) or issubdtype(dtype(type(value)), dtype(float))

    def __is_undefined__(value):
        return isnull(value)

    def __predicate__(value):
        if Encoder.__is_numeric__(value) and not Encoder.__is_undefined__(value):
            return operator.ge
        elif Encoder.__is_numeric__(value) and Encoder.__is_undefined__(value):
            return operator.eq
        else:
            return operator.eq

    def __init__(self, data, header=None, mode="complete", target=None):
        self.base = 10
        if len(data.shape) <= 1:
            data = array([ [data[i]] for i in range(data.shape[0]) ])
        (n, m) = data.shape

        if target is None:
            target = [i for i in range(n)]
        else:
            target = target.tolist()

        self.mode = mode
        self.values = [ SortedDict() for j in range(m)] # Track the uniquely occuring values for each column
        self.types = ['Numerical' for j in range(m)]
        self.optional = [True for j in range(m)]
        self.cardinalities = [ None for j in range(m)]
        self.defaults = [None for j in range(m)]
        self.magnitudes = [None for j in range(m)]
        self.radii = [None for j in range(m)]
        self.radial_values = [ SortedDict() for j in range(m)] # Track the uniquely occuring values for each column
        self.names = header

        if self.mode == "none":
            self.headers = header
            return

        if self.mode == "tree":
            from python.model.tree_encoder import TreeEncoder
            (n,m) = data.shape
            clf = RandomForestClassifier(n_estimators=10, max_features=None, random_state=0)
            clf.fit(data, array([ [i] for i in target ]))  
            clf.feature_importances_
            importance_index = argsort(clf.feature_importances_)[::-1]
            def key(sample):
                # print(tuple(sample[i] for i in importance_index))
                return tuple(sample[i] for i in importance_index)
            self.tree = TreeEncoder([ tuple(data[i,:]) for i in range(n) ], key)
            # print("Unique Samples: {}, Tree Width: {} ".format(len(set(tuple(data[i,:]) for i in range(n))), self.tree.width()))
            self.headers = [ "path[{}]".format(i) for i in range(self.tree.depth()) ]
            return

        for j in range(m):
            for i in range(n):
                value = data[i,j]
                if isnull(value) or value == '':
                    self.optional[j] = True
                elif Encoder.__is_numeric__(value):
                    self.types[j] = 'Numerical'
                    if not value in self.values[j]:
                        self.values[j][value] = set()
                    self.values[j][value].add(target[i])

                    # Radial Analysis
                    mantissa = float(value)
                    magnitude = 0
                    if mantissa != 0:
                        while round(mantissa) != mantissa:
                            mantissa = mantissa * self.base
                            magnitude += 1
                        while round(mantissa / self.base) == (mantissa / self.base):
                            mantissa = mantissa / self.base
                            magnitude -= 1

                    if self.magnitudes[j] is None:
                        self.magnitudes[j] = magnitude
                    else:
                        self.magnitudes[j] = max(self.magnitudes[j], magnitude)
                else:
                    self.types[j] = 'Categorical'
                    if not value in self.values[j]:
                        self.values[j][value] = set()
                    self.values[j][value].add(target[i])

            self.cardinalities[j] = len(self.values[j])
            if self.types[j] == 'Numerical' and len(self.values[j]) > 1:

                while len(set( int(value * pow(self.base, self.magnitudes[j])) for value in self.values[j].keys())) == len(set( int(value * pow(self.base, self.magnitudes[j]-1)) for value in self.values[j].keys())):
                    self.magnitudes[j] -= 1

                mantissa = int(max(self.values[j].keys()) * pow(self.base, self.magnitudes[j]))
                self.radii[j] = ceil(log(mantissa, self.base))
                values = list(sorted( value * pow(self.base, self.magnitudes[j]) for value in self.values[j].keys()))

                for k in range(self.radii[j]):
                    radial_values = SortedSet( int(value / pow(self.base, k)) % self.base for value in values)
                    if len(radial_values) > 1:
                        self.radial_values[j][k] = radial_values      

        self.encoders = [None for j in range(m)]

        for j in range(m):
            try:
                if self.cardinalities[j] <= 1 and self.optional == False:
                    self.types[j] = 'Redundant'
                elif self.cardinalities[j] <= 2:
                    self.types[j] = 'Binary'
                elif self.cardinalities[j] <= 5 and all(type(value) == type(0) for value in self.values[j]):
                    self.types[j] = 'Categorical'
                elif self.mode == "radix" and sum(len(values) for radial_index, values in self.radial_values[j].items()) < self.cardinalities[j]:
                    self.types[j] = 'Radial'

                if self.types[j] == 'Redundant':
                    self.encoders[j] = None
                elif self.types[j] == 'Binary':
                    if self.optional[j]:
                        self.encoders[j] = [
                            { 'relation': operator.eq, 'reference': 0, 'value': 0 },
                            { 'relation': operator.eq, 'reference': 1, 'value': 1 },
                        ]
                    else:
                        self.defaults[j] = min(self.values[j])
                        self.encoders[j] = [
                            { 'relation': operator.eq, 'reference': 1, 'value': 1 },
                        ]
                elif self.types[j] == 'Categorical':
                    if self.optional[j]:
                        self.encoders[j] = [ { 'relation': operator.eq, 'reference': value, 'value': value } for value in self.values[j] ]
                    else:
                        self.defaults[j] = min(self.values[j])
                        self.encoders[j] = [ { 'relation': operator.eq, 'reference': value, 'value': value } for value in list(self.values[j].keys())[1:] ]
                elif self.types[j] == 'Radial':
                    if self.optional[j]:
                        self.defaults[j] = min(self.values[j])
                    self.encoders[j] = []
                    for k in range(self.radii[j]):
                        if k in self.radial_values[j] and len(self.radial_values[j][k]) > 1:
                            for radial_value in self.radial_values[j][k]:
                                self.encoders[j].append({ 'relation': operator.ge, 'reference': radial_value, 'value': radial_value , "radial_index": k - self.magnitudes[j] })
                elif self.types[j] == "Numerical":
                    feature_values = list(self.values[j].keys())
                    base_value = feature_values[0]
                    references = [base_value]
                    for value in feature_values[1:]:
                        reference = (base_value + value) / 2.0
                        union = self.values[j][base_value].union(self.values[j][value])
                        if self.mode == "bucketize" and len(union) <= 1:
                            references.append(None)
                        else:
                            references.append(reference)
                        base_value = value
                    if self.optional[j]:
                        self.encoders[j] = [ { 'relation': operator.ge, 'reference': references[k], 'value': feature_values[k] } for k in range(0, len(self.values[j])) if not references[k] is None ]
                    else:
                        self.defaults[j] = min(feature_values)
                        self.encoders[j] = [ { 'relation': operator.ge, 'reference': references[k], 'value': feature_values[k] } for k in range(1, len(self.values[j])) if not references[k] is None ]
            except Exception as e:
                print("Exception({}) at Column {}".format(e, j))
                print("Cardinalities[j] = {}".format(self.cardinalities[j]))
                print("Optional[j] = {}".format(self.optional[j]))
                print("Type[j] = {}".format(self.types[j]))
                print("Values[j] = {}".format(self.values[j]))
                print([type(values) for values in self.values[j]])
                raise e

        # discrete = 0
        # continuous = 0
        # for j in range(m):
        #     if self.types[j] == 'Numerical':
        #         continuous += 1
        #     else:
        #         discrete += 1
        #     if not header is None:
        #         print("Index: {}, Name: {}, Type: {}, Cardinality: {}, Encoders: {}, Optional: {}, Radius: {}, Magnitude: {}".format(
        #             j, header[j], self.types[j], self.cardinalities[j], len(self.encoders[j]), self.optional[j], self.radii[j], self.magnitudes[j]))
        #     else:
        #         print("Index: {}, Name: {}, Type: {}, Cardinality: {}, Encoders: {}, Optional: {}, Radius: {}, Magnitude: {}".format(
        #             j, 'Unknown', self.types[j], self.cardinalities[j], len(self.encoders[j]), self.optional[j], self.radii[j], self.magnitudes[j]))
        # print("Number of Discrete Columns: {}, Number of Continuous Columns: {}, Total Cardinality: {}\n".format(discrete, continuous, sum(self.cardinalities)))

        offset = 0
        self.headers = []
        for j, encoders in enumerate(self.encoders):
            if encoders != None:
                if self.types[j] == "Radial":
                    self.headers += [ 
                        '{}{}{}'.format(
                            'x[i,{}][{}]'.format(j, encoder['radial_index']) if header is None else '{}[{}]'.format(header[j], encoder['radial_index']),
                            '>=' if encoder['relation'] == operator.ge else '==',
                            encoder['reference']
                        ) for encoder in encoders ]
                else:
                    self.headers += [ 
                        '{}{}{}'.format(
                            'x[i,{}]'.format(j) if header is None else header[j],
                            '>=' if encoder['relation'] == operator.ge else '==',
                            encoder['reference']
                        ) for encoder in encoders ]
        # print(self.headers)
    
    def encode(self, data):
        if self.mode == "none":
            return data

        if len(data.shape) <= 1:
            data = array([[data[i]] for i in range(data.shape[0])])
        (n, m) = data.shape
        encoded = []
        for i in range(n):
            if self.mode == "tree":
                # print(data[i,:], "=>", self.tree.encode(data[i,:]))
                encoded.append(self.tree.encode(data[i,:]))
                continue
            row = []                
            for j in range(m):
                encoders = self.encoders[j]
                if encoders == None:
                    continue
                cells = []
                for encoder in encoders:
                    if isnan(data[i, j]):
                        cell = 0
                    elif "radial_index" in encoder:
                        try:
                            value = int(data[i, j] * pow(self.base, -encoder["radial_index"])) % self.base
                        except:
                            print(data[i, j], pow(self.base, -encoder["radial_index"]))
                        cell = int(encoder['relation'](value, encoder['reference']))
                    else:
                        cell = int(encoder['relation'](data[i, j], encoder['reference']))
                    cells.append(cell)
                row += cells
            encoded.append(row)
        return array(encoded)

    def decode(self, data):
        if len(data.shape) <= 1:
            data = array([[data[i]] for i in range(data.shape[0])])
        (n, m) = data.shape
        decoded = []
        for i in range(n):
            offset = 0
            row = []
            for j, encoders in enumerate(self.encoders):
                # print(data[i, offset:offset+len(encoders)])
                if any(data[i, offset+k] == 1 for k in len(encoders)):
                    row.append(max(encoder['value'] for k, encoder in enumerate(encoders) if data[i, offset+k] == 1))
                else:
                    row.append(self.defaults[j])
                offset += len(encoders)
            decoded.append(row)
        return array(decoded)

    def decode_leaves(self, tree):
        if tree['leaf'] == True:
            # del tree['leaf']
            total = sum(weight for _, weight in tree['distribution'].items())
            distribution = {}
            for encoded_label, weight in tree['distribution'].items():
                bits = [int(c) for c in str(encoded_label)]
                bits.reverse() # Must reverse due to endianess
                decoded_label = self.defaults[0]
                encoders = self.encoders[0]
                if any(bits[k] == 1 for k in range(len(encoders))):
                    decoded_label = max(encoder['value'] for k, encoder in enumerate(encoders) if bits[k] == 1)
                if type(decoded_label) in { int, int64, int32 }:
                    distribution[int(decoded_label)] = weight / total
                elif type(decoded_label) in { float, float64, float32 }:
                    distribution[float(decoded_label)] = weight / total
                else:
                    distribution[str(decoded_label)] = weight / total
            tree['distribution'] = distribution
            if not self.names is None:
                tree['name'] = self.names[0]
        elif tree['leaf'] == False:
            for value, subtree in tree['branches'].items():
                self.decode_leaves(subtree)
        return tree

    def decode_branches(self, tree):
        if 'leaf' in tree and tree['leaf'] == False:
            for value, subtree in tree['branches'].items():
                self.decode_branches(subtree)
            # del tree['leaf']
            encoded_feature_index = tree['feature_index']
            offset = 0
            tree['branches'][str(False)] = tree['branches'][str(int(False))]
            tree['branches'][str(True)] = tree['branches'][str(int(True))]
            del tree['branches'][str(int(False))]
            del tree['branches'][str(int(True))]
            for j, encoders in enumerate(self.encoders):
                if encoded_feature_index < offset + len(encoders):
                    tree['feature_index'] = j
                    encoder = encoders[encoded_feature_index - offset]
                    tree['reference'] = encoder['reference']
                    if encoder['relation'] == operator.ge:
                        tree['relation'] = '>='
                    elif encoder['relation'] == operator.eq:
                        tree['relation'] = '=='
                    if not self.names is None:
                        tree['name'] = self.names[j]
                    break
                offset += len(encoders)
        return tree
                    



import numpy as np

class BaseEstimator:
    """
    Kaplan Meier / Nelson Aalen estimator
    """
    def __init__(self, event=None, y=None, reverse=False):
        """constructor of Kaplan Meier Estimator"""
        self.unique_times = None
        self.cumu_harzard_prob = None 
        self.survival_prob = None 
        if event is not None and y is not None:
            self.fit(event=event, y=y, reverse=reverse)

    def fit(self, event, y, reverse=False):
        """
        Parameters
        ---
        event: array-like, shape = [n_samples, ]
            an n-by-1 column of event associated with each sample
        y : array-like, shape = [n_samples, ]
            column containing the observed time for each sample in X
        reverse: bool
            fit a censoring distribution if reverse is True
        Modifies
        ---
        fit the Kaplan Meier / Nelson Aalen Estimator
        """
        n_samples = len(y)
        
        if reverse:
            event = 1 - event 
        order = np.argsort(y)

        time_threshold = np.zeros(n_samples, dtype=y.dtype)
        num_event_at_threshold = np.zeros(n_samples, dtype=int)
        cumu_num_total_at_threshold = np.zeros(n_samples, dtype=int)

        num_total = 0 
        num_event = 0 
        j = 0
        for i in range(n_samples):
            num_total += 1
            num_event += event[order[i]]
            if i < n_samples - 1 and y[order[i]] != y[order[i+1]]:
                 time_threshold[j] = y[order[i]]
                 num_event_at_threshold[j] = num_event
                 cumu_num_total_at_threshold[j] = num_total
                 num_event = 0 
                 j += 1
            elif i == n_samples - 1:
                time_threshold[j] = y[order[i]]
                num_event_at_threshold[j] = num_event
                cumu_num_total_at_threshold[j] = num_total
                j += 1
        
        time_threshold = np.resize(time_threshold, j)
        num_event_at_threshold = np.resize(num_event_at_threshold, j)
        cumu_num_total_at_threshold = np.resize(cumu_num_total_at_threshold, j)

        num_risky_at_threshold = n_samples - cumu_num_total_at_threshold
        num_risky_at_threshold = np.r_[n_samples, num_risky_at_threshold]
        num_risky_at_threshold = num_risky_at_threshold[:-1]

        ratio = num_event_at_threshold / num_risky_at_threshold # care for NaN here

        survival_prob = np.cumprod(1.0 - ratio)
        cumu_harzard_prob = np.cumprod(ratio)
        # take care before time before first time threshold 
        self.unique_times = np.r_[-np.infty, time_threshold]
        self.survival_prob = np.r_[1.0, survival_prob]
        self.cumu_harzard_prob = np.r_[0.0, cumu_harzard_prob]

        return self
    
    def predict_survival_prob(self, times):
        """
        Parameters
        ---
        times: array-like, shape = [n_times, ]
            an vector of time thresholds at which to predict survival probability
        ---
        Returns
        ---
            survival probabilities: array-like, shape = [n_times, ]
        """
        indices = np.searchsorted(self.unique_times, times, side="right") - 1

        return self.survival_prob[indices] 
    
    def predict_cumu_harzard_prob(self, times):
        """
        Parameters
        ---
        times: array-like, shape = [n_times, ]
            an vector of time thresholds at which to predict survival probability
        ---
        Returns
        ---
            cumulative harzard probability: array-like, shape = [n_times, ]
        """

        indices = np.searchsorted(self.unique_times, times, side="right") - 1

        return self.cumu_harzard_prob[indices] 


    def predict_ipcw(self, times):
        """
        Parameters
        ---
        times: array-like, shape = [n_times, ]
            an vector of time thresholds at which to predict survival probability
        ---
        Returns
        ---
            inverse probability of censoring weights: array-like, shape = [n_times, ]
        """
        probs = self.predict_survival_prob(times)
        probs[probs == 0] = np.inf
        return 1 / probs
        
    


            

            
            



import numpy as np
import pandas as pd
import json
import time
import random
import sys
import os  
from queue import Queue
import pathlib

from math import ceil
from sklearn.model_selection import KFold, cross_val_score, train_test_split
from sklearn.ensemble import GradientBoostingRegressor
from sklearn import metrics


# fit the tree using gradient boosted classifier
def fit_boosted_tree(X, y, n_est=10, lr=0.1, d=1):
    clf = GradientBoostingRegressor(loss='squared_error', learning_rate=lr, n_estimators=n_est, max_depth=d,
                                    random_state=42)
    clf.fit(X, y)
    out = clf.score(X, y)
    return clf, out


# perform cut on the dataset
def cut(X, ts):
    df = X.copy()
    colnames = X.columns
    for j in range(len(ts)):
        for s in range(len(ts[j])):
            X[colnames[j]+'<='+str(ts[j][s])] = 1
            k = df[colnames[j]] > ts[j][s]
            X.loc[k, colnames[j]+'<='+str(ts[j][s])] = 0
        X = X.drop(colnames[j], axis=1)
    return X


# compute the thresholds
def get_thresholds(X, y, n_est, lr, d, backselect=True):
    # got a complaint here...
    y = np.ravel(y)
    # X is a dataframe
    clf, out = fit_boosted_tree(X, y, n_est, lr, d)
    #print('acc:', out, 'acc cv:', score.mean())
    thresholds = []
    for j in range(X.shape[1]):
        tj = np.array([])
        for i in range(len(clf.estimators_)):
            f = clf.estimators_[i,0].tree_.feature
            t = clf.estimators_[i,0].tree_.threshold
            tj = np.append(tj, t[f==j])
        tj = np.unique(tj)
        thresholds.append(tj.tolist())

    X_new = cut(X, thresholds)
    clf1, out1 = fit_boosted_tree(X_new, y, n_est, lr, d)
    #print('acc','1:', out1, 'acc1 cv:', scorep.mean())

    outp = 1
    Xp = X_new.copy()
    clfp = clf1
    itr=0
    if backselect:
        while outp >= out1 and itr < X_new.shape[1]-1:
            vi = clfp.feature_importances_
            if vi.size > 0:
                c = Xp.columns
                i = np.argmin(vi)
                Xp = Xp.drop(c[i], axis=1)
                clfp, outp = fit_boosted_tree(Xp, y, n_est, lr, d)
                itr += 1
            else:
                break
        Xp[c[i]] = X_new[c[i]]
        #_, _ = fit_boosted_tree(Xp, y, n_est, lr, d)

    h = Xp.columns
    #print('features:', h)
    return Xp, thresholds, h

# compute the thresholds
def compute_thresholds(X, y, n_est, max_depth) :
    # n_est, max_depth: GBDT parameters
    # set LR to 0.1
    lr = 0.1
    start = time.perf_counter()
    X, thresholds, header = get_thresholds(X, y, n_est, lr, max_depth, backselect=True)
    guess_time = time.perf_counter()-start

    return X, thresholds, header, guess_time











import json
import osst.libosst as osst  # Import the OSST extension
from osst.model.tree_survival_regressor import TreeSurvivalRegressor  # Import the tree survival regressor model

class OSST:
    def __init__(self, configuration={}):
        self.configuration = configuration
        self.time = 0.0
        # self.stime = 0.0
        # self.utime = 0.0
        # self.maxmem = 0
        # self.numswap = 0
        # self.numctxtswitch = 0
        self.iterations = 0
        self.size = 0
        self.tree = None
        self.encoder = None
        self.lb = 0
        self.ub = 0
        self.timeout = False
        self.reported_loss = 0

    def load(self, path):
        """
        Parameters
        ---
        path : string
            path to a JSON file representing a model
        """
        with open(path, 'r') as model_source:
            result = model_source.read()
        result = json.loads(result)
        self.tree = TreeSurvivalRegressor(result[0])

    def __train__(self, X, event, y):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            matrix containing the training samples and features
        event: array-like, shape = [n_samples by 1]
            an n-by-1 column of event associated with each sample
        y : array-like, shape = [n_samples by 1]
            column containing the correct label for each sample in X
        Modifies
        ---
        trains a model using the OSST native extension
        """
        (n, m) = X.shape
        dataset = X.copy()
        dataset.insert(m, "event", event)
        dataset.insert(m + 1, "time", y)  # It is expected that the last column is the label column

        osst.configure(json.dumps(self.configuration, separators=(',', ':')))
        result = osst.fit(dataset.to_csv(index=False))  # Perform extension call to train the model

        self.time = osst.time()  # Record the training time

        if osst.status() == 0:
            print("osst reported successful execution")
            self.timeout = False
        elif osst.status() == 2:
            print("osst reported possible timeout.")
            self.timeout = True
            self.time = -1
            # self.stime = -1
            # self.utime = -1
        else:
            print('----------------------------------------------')
            print(result)
            print('----------------------------------------------')
            raise Exception("Error: OSST encountered an error while training")

        result = json.loads(result)  # Deserialize resu

        self.tree = TreeSurvivalRegressor(result[0], X=X, event=event, y=y)  # Parse the first result into model
        self.iterations = osst.iterations()  # Record the number of iterations
        self.size = osst.size()  # Record the graph size required

        # self.maxmem = osst.maxmem()
        # self.numswap = osst.numswap()
        # self.numctxtswitch = osst.numctxtswitch()

        self.lb = osst.lower_bound()  # Record reported global lower bound of algorithm
        self.ub = osst.upper_bound()  # Record reported global upper bound of algorithm
        self.reported_loss = osst.model_loss()  # Record reported training loss of returned tree

        print("training completed. {:.3f} seconds.".format(self.time))
        print("bounds: [{:.6f}..{:.6f}] ({:.6f}) IBS loss = {:.6f}, iterations={}".format(self.lb, self.ub, self.ub - self.lb,
                                                                                    self.reported_loss,
                                                                                    self.iterations))

    def fit(self, X, event, y):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            matrix containing the training samples and features
        event: array-like, shape = [n_samples by 1]
            an n-by-1 column of event associated with each sample
        y : array-like, shape = [n_samples by 1]
            column containing the correct label for each sample in X
        Modifies
        ---
        trains the model so that this model instance is ready for prediction
        """

        self.__train__(X, event, y)

        return self

    def predict(self, X):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            a matrix where each row is a sample to be predicted and each column is a feature to be used for prediction
        Returns
        ---
        array-like, shape = [n_sampels by 1] : a column where each element is the leaf node index associated with each row
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.predict(X)

    def score(self, X, event, y):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            an n-by-m matrix of sample and their features
        event: array-like, shape = [n_samples by 1]
            an n-by-1 column of event associated with each sample
        y : array-like, shape = [n_samples by 1]
            an n-by-1 column of labels associated with each sample
            
        Returns
        ---
        real number : IBS score
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.score(X, event, y)

    def __len__(self):
        """
        Returns
        ---
        natural number : The number of terminal nodes present in this tree
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return len(self.tree)

    def leaves(self):
        """
        Returns
        ---
        natural number : The number of terminal nodes present in this tree
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.leaves()

    def nodes(self):
        """
        Returns
        ---
        natural number : The number of nodes present in this tree
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.nodes()

    def max_depth(self):
        """
        Returns
        ---
        natural number : the length of the longest decision path in this tree. A single-node tree will return 1.
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.maximum_depth()

    def latex(self):
        """
        Note
        ---
        This method doesn't work well for label headers that contain underscores due to underscore being a reserved character in LaTeX
        Returns
        ---
        string : A LaTeX string representing the model
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.latex()

    def json(self):
        """
        Returns
        ---
        string : A JSON string representing the model
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.json()

    def predict_survival_function(self, X):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            a matrix where each row is a sample to be predicted and each column is a feature to be used for prediction
        Returns
        ---
        survival functions: array-like, shape = [n_sampels, ]
             a column where each element is the predicted survival function associated with each row
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.predict_survival_function(X)
    
    def predict_cumulative_harzard_function(self, X):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            a matrix where each row is a sample to be predicted and each column is a feature to be used for prediction
        Returns
        ---
        cumulative hazard function: array-like, shape = [n_sampels, ]
            a column where each element is the predicted cumulative harzard function associated with each row
        """
        if self.tree is None:
            raise Exception("Error: Model not yet trained")
        return self.tree.predict_cumulative_harzard_function(X)


import platform
import os
import pathlib
import distro

from setuptools import find_packages
from skbuild import setup

cmake_args = []

if platform.system() == "Windows" or (platform.system() == "Linux" and distro.id() == "centos"):
    assert "VCPKG_INSTALLATION_ROOT" in os.environ, \
        "The environment variable \"VCPKG_INSTALLATION_ROOT\" must be set before running this script."
    toolchain_path = pathlib.Path(os.getenv("VCPKG_INSTALLATION_ROOT")) / "scripts/buildsystems/vcpkg.cmake"
    cmake_args.append("-DCMAKE_TOOLCHAIN_FILE={}".format(toolchain_path))

print("Additional CMake Arguments = {}".format(cmake_args))

setup(
    name="osst",
    version="0.1.4",
    description="Implementation of Optimal Sparse Survival Trees",
    author="UBC Systopia Research Lab",
    url="https://github.com/ruizhang1996/optimal-sparse-survival-trees-public",
    license="BSD 3-Clause",
    packages=find_packages(where='.'),
    cmake_install_dir="osst",
    cmake_args=cmake_args,
    python_requires=">=3.9",
    long_description=pathlib.Path("README_PyPI.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    install_requires=["setuptools",
                      "wheel",
                      "attrs",
                      "packaging>=20.9",
                      "editables==0.2;python_version>'3.0'",
                      "pandas",
                      "scikit-learn",
                      "sortedcontainers",
                      "gmpy2",
                      "matplotlib",
                      "scikit-survival"
                    ]
)
